import random


def generate(seed):
	lst = {i: 0 for i in range(256)}
	used = []
	random.seed(seed)

	def g():
		r = random.randint(0,255)
		while r in used:
			r = random.randint(0,255)
		used.append(r)
		return r

	for i in range(128):
		gen = g()
		gen2 = g()
		lst[gen] = gen2
		lst[gen2] = gen

	return lst

def generate_configuration(seed, rotors=3):
	configuration = {'rotors': [generate(seed+bytes(i)) for i in range(rotors)], 'commutators': generate(seed+bytes(3))}
	return configuration


class Enigma:
	# TODO: add rotors move
	"""
		uses library random
		example:
		from egter.crypt import enigma
		e = enigma.Enigma(enigma.generate_configuration(b'pass'), b'pass')
		enc = e.encode(b'hello')
		dec = e.decode(enc)
		print(enc, dec)
		# b'\\xa7G&\\x1cP' b'hello'
	"""

	def __init__(self, configuration, seed, state=None, save_state=False):
		self.configuration = configuration
		# self.save_state = save_state
		# self.state = state
		self.seed = seed
		# if not state:
		# 	self.state = [0 for i in configuration['rotors']]
		# self.old_state = list(self.state)
		self.rotors_len = len(configuration['rotors'])

	# def rotor_down(self):
	# 	states = []
	# 	next_pl = False
	# 	self.state[0] += 1
	# 	for state in self.state:
	# 		if next_pl:
	# 			state += 1
	# 		if state > 255:
	# 			next_pl = True
	# 			state -= 256
	# 		else:
	# 			next_pl = False
	# 		states.append(state)
	# 	self.state = states

	# def rotor_up(self):
	# 	states = []
	# 	next_pl = False
	# 	self.state[0] -= 1
	# 	for state in self.state:
	# 		if next_pl:
	# 			state -= 1
	# 		if state < 0:
	# 			next_pl = True
	# 			state += 256
	# 		else:
	# 			next_pl = False
	# 		states.append(state)
	# 	self.state = states

	def get_byte(self, byte):
		tmp = byte
		for rotor in range(self.rotors_len):
			tmp = self.configuration['commutators'][tmp]
			tmp = self.configuration['rotors'][rotor][tmp]
			tmp = self.configuration['commutators'][tmp]
		lst = list(range(self.rotors_len))
		lst.reverse()
		# for rotor in lst:
		# 	tmp = self.configuration['commutators'][tmp]
		# 	print('->', 'rotor:', rotor, 'byte:', tmp, 'byte2:', self.configuration['rotors'][rotor][tmp])
		# 	tmp = self.configuration['rotors'][rotor][tmp]
		# 	tmp = self.configuration['commutators'][tmp]
		tmp += random.randint(0,255)
		if tmp > 255:
			tmp -= 256
		return tmp

	def get_byte_reverse(self, byte):
		tmp = byte
		tmp -= random.randint(0,255)
		if tmp < 0:
			tmp += 256
		lst = list(range(self.rotors_len))
		lst.reverse()
		for rotor in lst:
			tmp = self.configuration['commutators'][tmp]
			tmp = self.configuration['rotors'][rotor][tmp]
			tmp = self.configuration['commutators'][tmp]
		# for rotor in range(self.rotors_len):
		# 	tmp = self.configuration['commutators'][tmp]
		# 	print('->', 'rotor:', rotor, 'byte:', tmp, 'byte2:', self.configuration['rotors'][rotor][tmp])
		# 	tmp = self.configuration['rotors'][rotor][tmp]
		# 	tmp = self.configuration['commutators'][tmp]
		return tmp

	def encode(self, bt):
		random.seed(self.seed)
		bt = list(bytearray(bt))
		bts = []
		for byte in bt:
			byte = self.get_byte(byte)
			bts.append(byte)
		# if self.save_state == False:
		# 	self.state = self.old_state
		return bytes(bts)

	def decode(self, bt):
		random.seed(self.seed)
		bt = list(bytearray(bt))
		bts = []
		for byte in bt:
			byte = self.get_byte_reverse(byte)
			bts.append(byte)
		# if self.save_state == False:
		# 	self.state = self.old_state
		return bytes(bts)


class List(list):
    """
    usage:

    from egter import customs
    List = customs.List
    lst = [0, 1, 4]
    print(lst.get(3, None))
    # None
    print(lst.get(2, None))
    # 4
    print(lst.get(3))
    # None

    equal to:
    d = {0: 0, 1: 1, 2: 4}
    print(d.get(4)
    # None
    """

    def get(self, i, ret=None):
        try:
            return self[i]
        except:
            return ret


import random

class Stenography:
    """
    import egter_test.crypt as crypt
    import egter_test.crypt.defaults as defaults

    s = crypt.Stenography(defaults.stenography_list)
    enc = s.encode(b'some bytes')
    dec = s.decode(enc)
    print(enc, dec)
    # Viverra facilisisu eue nisla elita nisio luctuso litorae, nisla; viverra b'some bytes'
    """

    def __init__(self, stenography_list, punctuation=',:;!?.', punctuation_end='!?.'):
        self.stenography_list = stenography_list
        self.punctuation = punctuation
        self.punctuation_end = punctuation_end

    def encode(self, bt):
        out = ''
        pl = 0
        up = True
        for i in range(len(bt)):
            try:
                f = str(bt[i + pl])
                try:
                    s = str(bt[i + 1 + pl])
                    n = f + ':' + s
                except:
                    s = False
                    n = False
                if s and random.choice([True, True, False]):
                    if n in self.stenography_list:
                        pl += 1
                        if up:
                            out += self.stenography_list[n].title()
                            up = False
                        else:
                            out += self.stenography_list[n]
                    else:
                        if up:
                            out += self.stenography_list[f].title()
                            up = False
                        else:
                            out += self.stenography_list[f]
                else:
                    if up:
                        out += self.stenography_list[f].title()
                        up = False
                    else:
                        out += self.stenography_list[f]
                if random.choice([True, False, False, False, False]):
                    punc = random.choice(self.punctuation)
                    out += punc
                    if punc in self.punctuation_end:
                        up = True
                out += ' '
            except:
                pass
        out = out.strip()
        return out

    def decode(self, out):
        out = out.lower()
        for i in self.punctuation:
            out = out.replace(i, '')
        spl = out.split()
        bt = []
        for word in spl:
            for i in self.stenography_list:
                if self.stenography_list[i] == word:
                    s = i.split(':')
                    for k in s:
                        bt.append(int(k))
        bt = bytes(bt)
        return bt


from distutils.core import setup

setup(
    name='egter',
    version='0.3',
    packages=['egter', 'egter.crypt', 'egter.crypt.enigma', 'egter.crypt.steganography', 
    			'egter.crypt.hash'],
    url='https://notabug.org/EgTer/egter-py',
    license='GNU GPL v3',
    author='EgTer',
    author_email='annom2017@mail.ru',
    description='My collection of my tools'
)



import random
import binascii


class EHash:
    """
    (C) Egor Ternovoy. 2020

    !!! NOT SO SAFE! It is possible to reverse calculate source bytes from hash! !!!

    Uses library random

    Usage:
        from egter import crypt

        hash = crypt.EHash(b'some bytes', lenght=32, iterations=100)  # lenght - bytes count

        print(hash) # equal to hexdigest method
        # 90b0f287805156865bfebbf93eac3328ab0e9bcbb82dfcc9d5f88d6d41524766

        print(hash.digest()) # bytes
        # b'\x90\xb0\xf2\x87\x80QV\x86[\xfe\xbb\xf9>\xac3(\xab\x0e\x9b\xcb\xb8-\xfc\xc9\xd5\xf8\x8dmARGf'

        print(hash.hexdigest())
        # 90b0f287805156865bfebbf93eac3328ab0e9bcbb82dfcc9d5f88d6d41524766

        print(hash.intdigest())
        # 65445689155113512790458496256639542022557958647760937310771143133228635932518

        hash.update(b'add bytes to hash')
        print(hash.hexdigest())
        # 591822f89a114ede35c8ac4a4ebeda6952fd6f4114a9434ebe27b1b8fc6b0d2a

        # set custom seed
        hash.generate(b'some seed')
        print(hash.hexdigest())
        # a2c401e4e18b0ebda141fd63da4a0efbc9253b9d204e6c24381b9bd45ec4abc1

        hash2 = crypt.EHash(int) # equal to crypt.EHash('<class 'int'>')
        print(hash)
        # a7eb7705b8a37e77df4a48882584b4bcfa20377b305bdae586784636be9ba255

        custom_hash = crypt.EHash('custom hash', custom=(1, 2, 3), custom_type=str) # custom_type = str|bytes,
                                                                                           #custom = iterable type
        print(custom_hash.customdigest())
        # 11321321121122112121111133122213
    """

    def __init__(self, text, lenght=32, iterations=100, custom=(1, 2, 3), custom_type=str):
        """Creates EHash object"""
        if not isinstance(text, (str, bytes, bytearray)):
            text = str(text)
        if isinstance(text, str):
            text = text.encode()
        self.custom = custom
        self.custom_type = custom_type
        self.bytes = b''
        self.lenght = lenght
        self.iterations = iterations
        for i in range(iterations):
            for letter in text:
                self.generate(letter)

    def generate(self, seed):
        """Generating hash"""
        random.seed(self.bytes + bytearray(seed))
        self.bytes = bytes(random.choices(range(256), k=self.lenght))

    def update(self, text):
        """Update existing hash"""
        if not isinstance(text, (str, bytes, bytearray)):
            text = str(text)
        if isinstance(text, str):
            text = text.encode()
        for i in range(self.iterations):
            for letter in text:
                self.generate(letter)

    def digest(self):
        """Return BYTES"""
        return self.bytes

    def hexdigest(self):
        """Return HEX"""
        return binascii.hexlify(self.bytes).decode()

    def intdigest(self):
        """return INT"""
        return int.from_bytes(self.bytes, byteorder='big')

    def customdigest(self, custom=None):
        """return custom letters hash"""
        if custom == None:
            custom = self.custom
        return self.custom_type('').join([self.custom_type(i) for i in random.choices(custom, k=self.lenght)])

    def __str__(self):
        return binascii.hexlify(self.bytes).decode()

